///////////////////////////////////////////////////////////////////////////////
// Featured UART by J. Arias (2023)
// - Selectable Baud Rate, parity, and number of STOP bits
// - Data formats: 8N1, 8N2, 8E1, 8E2, 8O1, 8O2 
// - Buffered receiver (1 byte). Unbuffered transmitter
// - 99 Logic cells (DIVBITS = 8, ICE40HX FPGA)
///////////////////////////////////////////////////////////////////////////////
module UART_enh_core (
			// Parallel interface
	input	clk,		// System clock
	input	[7:0]d,		// Input bus for TX
	input	wr,			// write strobe (1 cycle): starts TX
	input	rd,			// read strobe (1 cycle): clears RX flags
	output	[7:0]q,		// Output bus for RX
			// Flags
	output	txrdy,		// TX ready flag
	output	rxvalid,	// RX valid data flag
	output	rxoverr,	// RX overrun error flag
	output	rxframeer,	// RX framing error flag
	output	parerr,		// RX parity error flag
			// Control
	input   nstop,		// Set to 1 for two stop bits
	input	paren,		// Set to 1 for parity
	input	parodd,		// Select odd parity if 1, even parity if 0
	input   [DIVBITS-1:0]bauddiv,	// baud rate divider baud=f_clk/(1+bauddiv)
			// Serial interface
	input 	rxd,		// serial data input
	output 	txd			// serial data output
);
parameter DIVBITS = 8;	// number of bits for clock counters

///////////////////// UART TX ////////////////////
reg [DIVBITS-1:0]txdiv=0;
wire txdivmax=(txdiv==bauddiv);
always @(posedge clk) 
	if (wr | txdivmax) txdiv<=0; // Reset if max or on writes
	else txdiv<=txdiv+1;

reg [9:0]txsh=10'h1FF;		// 10-bit shift register
wire txpar = ^{parodd,d}; 	// data parity
always @(posedge clk)
	if (wr) begin 
		`ifdef SIMULATION
	        $write ("%c",d); $fflush ( );
		`endif
		 txsh<={txpar,d,1'b0}; // Load data plus START bit
	end	else if (txdivmax) txsh<={1'b1,txsh[9]|(~paren),txsh[8:1]}; // shift 1'b1 in for future STOP bit
assign txd = txsh[0];

reg [3:0]txbitcnt=0;	// Bit counter (0000 means idle)
wire utxbusy=|txbitcnt;
// ntx = 10, 11, or 12 bits depending on parity and nstop
wire [3:0]ntx = {1'b1, paren&nstop, ~(paren&nstop), paren^nstop};

always @(posedge clk) 
	if (wr) txbitcnt<=ntx; 
	else if (utxbusy&txdivmax) txbitcnt<=txbitcnt-1;
	
assign txrdy=~utxbusy;
	
///////////////////// UART RX ////////////////////
reg [1:0]rxreg=0; // two samples of RXD
always @(posedge clk) rxreg<={rxreg[0],rxd};
reg [DIVBITS-1:0]rxdiv=0;
always @(posedge clk)
	if ((rxreg[1]^rxreg[0])|(rxdiv==bauddiv)) rxdiv<=0; // Reset if max or on any RXD edge
	else rxdiv<=rxdiv+1;
wire rxsample = (rxdiv=={1'b0,bauddiv[DIVBITS-1:1]}); // sample at the middle of the bit

reg [10:0]urxsh=11'h7FF; // 11-bit shift register
reg [7:0]urxbuffer;		// data buffer
reg rxframeer;			// frame error
reg parerr;				// parity error
reg [1:0]urxval=0;		// received data flags (00: no data, 01: data available, 11: overrun)
always @(posedge clk) begin
	if (~urxsh[0]) urxsh<=11'h7FF;  // if START bit at LSB set to all ones and copy results
	else if (rxsample) urxsh<={rxreg[0], paren?urxsh[10]:rxreg[0] ,urxsh[9:1]};
	
	if (~urxsh[0]) begin
		urxbuffer<=urxsh[8:1]; 	 // store received data
		rxframeer<=~urxsh[10];   // STOP bit in 0 means framing error
		parerr<= paren&(^{parodd,urxsh[9:1]});	// check parity
	end
	if (~urxsh[0]) urxval<={urxval[0],1'b1};  // set data valid and (maybe) overrun
	else if (rd) urxval<=0;  // clear data valid and overrun flags on reads
	
end
assign q = urxbuffer[7:0];	
assign rxvalid = urxval[0];
assign rxoverr = urxval[1];

endmodule

